use crate::runtime::{
    Reference,
    acquire, release,
    type_inner, type_key,
    block_length,
    block_set, block_get,
    block_set_natural, block_get_natural,
    block_set_integer, block_get_integer,
    block_set_field, block_get_field,
};
use crate::tag;
use super::block;

pub struct Block {
    managed:bool,
    addr:Reference,
}
impl Block {
    pub fn new(size:usize) -> Self
    {
        Self {
            managed:true,
            addr:unsafe {acquire(block(size))},
        }
    }

    pub fn with(size:usize, data:Vec<u8>) -> Self
    {
        let mut obj = Self::new(size);
        obj.set(data);
        return obj;
    }

    pub fn size(&self) -> usize
    {
        unsafe {block_length(self.addr)}
    }

    pub fn set(&mut self, data:Vec<u8>)
    {
        let length = unsafe {type_key(type_inner(self.addr.class))};
        for index in 0..usize::min(data.len(), length) {
            unsafe {block_set(self.addr, index, data[index])};
        }
    }

    pub fn set_u64(&self, data:u64)
    {
        unsafe {block_set_natural(self.addr, data)}
    }
    
    pub fn set_i64(&self, data:i64)
    {
        unsafe {block_set_integer(self.addr, data)}
    }

    pub fn set_field(&mut self, index:usize, length:usize, data:u64)
    {
        unsafe { block_set_field(self.addr, index, length, data); }
    }

    pub fn get(&self) -> Vec<u8>
    {
        let mut result = Vec::<u8>::new();
        let length = unsafe {type_key(type_inner(self.addr.class))};
        if length > 0 {
            result.resize(length, 0);
            for index in 0..length {
                result[index] = unsafe {block_get(self.addr, index)};
            }
        }
        return result;
    }

    pub fn get_u64(&self) -> u64
    {
        unsafe {block_get_natural(self.addr)}
    }
    
    pub fn get_i64(&self) -> i64
    {
        unsafe {block_get_integer(self.addr)}
    }

    pub fn get_field(&self, index:usize, length:usize) -> u64
    {
        unsafe {block_get_field(self.addr, index, length)}
    }
}
impl TryFrom<Reference> for Block {
    type Error = ();
    fn try_from(addr:Reference) -> Result<Self, Self::Error> {
        return if(unsafe {type_key(addr.class)} == tag::BLOCK) {
            Ok(Self { managed:false, addr:addr })
        }
        else {
            Err(())
        }
    }
}
impl std::ops::Deref for Block {
    type Target = Reference;
    fn deref(&self) -> &Self::Target { return &self.addr; }
}
impl Drop for Block {
    fn drop(&mut self) { if self.managed { unsafe {release(self.addr)}; } }
}
