use crate::runtime::{
    Reference,
    acquire, release,
    type_key,
    decimal_get, decimal_set, type_innerkey,
};
use crate::tag;
use super::decimal;

pub struct Decimal {
    managed:bool,
    addr:Reference,
}
impl Decimal {
    pub fn new(mantissa:usize) -> Self
    {
        Self {
            managed:true,
            addr:unsafe {acquire(decimal(mantissa))},
        }
    }

    pub fn with_raw(mantissa:usize, data:i64) -> Self
    {
        let mut obj = Self::new(mantissa);
        obj.set_raw(data);
        return obj;
    }

    pub fn with(mantissa:usize, value:f64) -> Self
    {
        let mut obj = Self::new(mantissa);
        obj.set(value);
        return obj;
    }

    pub fn copy(&self) -> Self
    {
        Decimal::try_from(self.addr).unwrap()
    }

    pub fn set_raw(&mut self, data:i64)
    {
        unsafe { decimal_set(self.addr, data) };
    }

    pub fn set(&mut self, value:f64)
    {
        let mantissa = unsafe {type_innerkey(self.addr.class)};
        let whole = value as i64;
        let decimal = ((value - value.trunc()).abs() * (1i64 << mantissa) as f64).floor() as i64;
        let data = (whole << mantissa) + (decimal as i64);
        self.set_raw(data);
    }

    pub fn get_raw(&self) -> i64
    {
        unsafe {decimal_get(self.addr)}
    }

    pub fn get(&self) -> f64
    {
        let mantissa = unsafe {type_innerkey(self.addr.class)};
        let data = self.get_raw();
        (data >> mantissa) as f64 + ((data & ((1i64 << mantissa) - 1)) as f64 / (1i64 << mantissa) as f64)
    }

    fn rebase(to:usize, from:usize, data:i64) -> i64
    {
        return if to > from {
            data << (to - from) as i64
        } else if to < from {
            data >> (from - to) as i64
        } else { data }
    }

    pub fn to_base(&self, mantissa:usize) -> Self
    {
        let mut result = Decimal::new(mantissa);
        result.set_raw(Self::rebase(mantissa, self.mantissa(), self.get_raw()));
        return result;
    }

    pub fn mantissa(&self) -> usize
    {
        unsafe {type_innerkey(self.addr.class)}
    }
}
impl Clone for Decimal {
    fn clone(&self) -> Self {
        Decimal::with_raw(self.mantissa(), self.get_raw())
    }
}
impl std::ops::Add for Decimal {
    type Output = Self;
    fn add(self, rhs: Self) -> Self::Output {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        Decimal::with_raw(lhe, lhd + rhd)
    }
}
impl std::ops::AddAssign for Decimal {
    fn add_assign(&mut self, rhs: Self) {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        self.set_raw(lhd + rhd);
    }
}
impl std::ops::Sub for Decimal {
    type Output = Self;
    fn sub(self, rhs: Self) -> Self::Output {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        Decimal::with_raw(lhe, lhd - rhd)
    }
}
impl std::ops::SubAssign for Decimal {
    fn sub_assign(&mut self, rhs: Self) {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        self.set_raw(lhd - rhd);
    }
}
impl std::ops::Mul for Decimal {
    type Output = Self;
    fn mul(self, rhs: Self) -> Self::Output {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        Decimal::with_raw(lhe, ((lhd as i128 * rhd as i128) >> lhe) as i64)
    }
}
impl std::ops::MulAssign for Decimal {
    fn mul_assign(&mut self, rhs: Self) {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        self.set_raw(((lhd as i128 * rhd as i128) >> lhe) as i64);
    }
}
impl std::ops::Div for Decimal {
    type Output = Self;
    fn div(self, rhs: Self) -> Self::Output {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        Decimal::with_raw(lhe, ((lhd as i128 * (1i64 << lhe) as i128) / rhd as i128) as i64)
    }
}
impl std::ops::DivAssign for Decimal {
    fn div_assign(&mut self, rhs: Self) {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        self.set_raw(((lhd as i128 * (1i64 << lhe) as i128) / rhd as i128) as i64);
    }
}
impl std::ops::Rem for Decimal {
    type Output = Self;
    fn rem(self, rhs: Self) -> Self::Output {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        Decimal::with_raw(lhe, (lhd as i128 % rhd as i128) as i64)
    }
}
impl std::ops::RemAssign for Decimal {
    fn rem_assign(&mut self, rhs: Self) {
        let lhe = unsafe {type_innerkey(self.addr.class)};
        let rhe = unsafe {type_innerkey(rhs.addr.class)};
        let lhd = self.get_raw();
        let rhd = Self::rebase(lhe, rhe, rhs.get_raw());

        self.set_raw(((lhd as i128 * (1i64 << lhe) as i128) % rhd as i128) as i64);
    }
}
impl std::ops::Neg for Decimal {
    type Output = Self;
    fn neg(self) -> Self::Output {
        let mantissa = unsafe {type_innerkey(self.addr.class)};
        Decimal::with_raw(mantissa, -self.get_raw())
    }
}
impl std::ops::Deref for Decimal {
    type Target = Reference;
    fn deref(&self) -> &Self::Target { return &self.addr; }
}
impl Drop for Decimal {
    fn drop(&mut self) { if self.managed { unsafe {release(self.addr)}; } }
}
impl TryFrom<Reference> for Decimal {
    type Error = ();
    fn try_from(addr:Reference) -> Result<Self, Self::Error> {
        return if(unsafe {type_key(addr.class)} == tag::DECIMAL) {
            Ok(Self { managed:false, addr:addr })
        }
        else {
            Err(())
        }
    }
}
impl Into<f64> for Decimal {
    fn into(self) -> f64 {
        self.get()
    }
}
