use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer,
        clone,
        block_set, block_get,
    },
};

pub struct Block<const N :usize> {
    address:Ref,
}
impl<const N :usize> Block<N> {
    pub fn new() -> Self
    {
        Self { address:unsafe {acquire(Self::type_id())} }
    }

    pub fn with(mut self, data:&[u8; N]) -> Self
    {
        self.set(data);
        self
    }

    pub fn set(&mut self, data:&[u8; N])
    {
        for i in 0..data.len() {
            unsafe {block_set(self.address.clone(), i, data[i])};
        }
    }

    pub fn get(&self) -> [u8; N]
    {
        let mut data = [0; N];
        for i in 0..data.len() {
            data[i] = unsafe {block_get(self.address.clone(), i)};
        }
        data
    }

    pub fn set_at(&mut self, index:usize, data:u8)
    {
        unsafe {block_set(self.address.clone(), index, data)}
    }

    pub fn get_at(&self, index:usize) -> u8
    {
        unsafe {block_get(self.address.clone(), index)}
    }

    pub fn set_unsigned<T: num::Num + num::NumCast>(&mut self, data:T)
    {
        let mut value :u128 = num::NumCast::from(data).unwrap();
        let n = N - 1;
        for i in 0..N {
            unsafe {block_set(self.address.clone(), n - i, (value & 0xFF) as u8)};
            value >>= 8;
        }
    }

    pub fn get_unsigned<T: num::Num + num::NumCast>(&mut self) -> T
    {
        let mut value :u128 = 0;
        for i in 0..N {
            value <<= 8;
            value |= unsafe {block_get(self.address.clone(), i)} as u128;
        }
        num::NumCast::from(value).unwrap()
    }

    pub fn len(&self) -> usize
    {
        N
    }
}
impl<const N :usize> SzunType for Block<N> {
    fn type_id() -> usize {
        let inner = unsafe {type_outer(0, N)};
        unsafe {type_outer(inner, types::TYPE_BLOCK)}
    }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}

pub fn block<const N :usize>(data:&[u8; N]) -> Block<N>
{
    Block::new().with(data)
}

impl<const N :usize> Clone for Block<N> {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())} }
    }
}

impl<const N :usize> TryFrom<Ref> for Block<N> {
    type Error = SzunError;

    fn try_from(addr: Ref) -> Result<Self, Self::Error> {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<const N :usize> Into<Ref> for Block<N> { fn into(self) -> Ref { self.address.clone() } }

impl<const N :usize> Drop for Block<N> {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
