use std::marker::PhantomData;
use crate::{
    prelude::*, runtime::{
        acquire, release, clone,
        type_outer, type_key,
        list_reserve, list_clear,
        list_insert, list_push, list_pop,
        list_at, list_back, list_set,
        list_length, list_capacity,
    }, types
};

pub struct List<T: SzunType> {
    address:Ref,
    _data:PhantomData<T>,
}
impl<T: SzunType> List<T> {
    pub fn new() -> Self {
        Self {
            address:unsafe {acquire(Self::type_id())},
            _data:PhantomData
        }
    }

    pub fn insert<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if index <= self.len() && self.compatible(data) {
            unsafe {list_insert(self.address.clone(), index, data.__to_ref())};
        }
    }

    pub fn push<D: SzunType>(&mut self, data:&D)
    {
        if self.compatible(data) {
            unsafe {list_push(self.address.clone(), data.__to_ref())};
        }
    }

    pub fn pop(&mut self) -> Option<T>
    {
        let refr = unsafe {list_back(self.address.clone())};
        if refr.is_some() {
            let result = Some(T::__from_ref(unsafe {clone(refr)}));
            unsafe {list_pop(self.address.clone())};
            result
        } else { None }
    }

    pub fn set<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {list_set(self.address.clone(), index, data.__to_ref())};
        }
    }

    pub fn clear(&mut self)
    {
        unsafe {list_clear(self.address.clone())};
    }

    pub fn reserve(&mut self, length:usize)
    {
        unsafe {list_reserve(self.address.clone(), length)};
    }

    pub fn get(&mut self, index:usize) -> T
    {
        let refr = unsafe {list_at(self.address.clone(), index)};
        if refr.is_some() {
            T::__from_ref(refr)
        } else { panic!("index out of bounds: {} / {}", index, self.len()) }
    }

    pub fn try_get(&mut self, index:usize) -> Option<T>
    {
        let refr = unsafe {list_at(self.address.clone(), index)};
        if refr.is_some() {
            Some(T::__from_ref(refr))
        } else { None }
    }

    pub fn len(&self) -> usize
    {
        unsafe {list_length(self.address.clone())}
    }

    pub fn capacity(&self) -> usize
    {
        unsafe {list_capacity(self.address.clone())}
    }
}
impl<T: SzunType> SzunType for List<T> {
    fn type_id() -> usize {
        unsafe {type_outer(T::type_id(), types::TYPE_LIST)}
    }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<R: SzunType>(&self, _:&R) -> bool {
        R::type_id() == T::type_id() || unsafe {type_key(T::type_id())} == types::TYPE_VARYING
    }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(address:Ref) -> Self { Self { address:address, _data:PhantomData } }
}
impl<T: SzunType> Clone for List<T> {
    fn clone(&self) -> Self {
        Self::new()
    }
}

impl<T: SzunType> TryFrom<Ref> for List<T> {
    type Error = SzunError;

    fn try_from(addr: Ref) -> Result<Self, Self::Error> {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr, _data:PhantomData })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<T: SzunType> Into<Ref> for List<T> {
    fn into(self) -> Ref { self.address.clone() }
}

impl<T: SzunType> Drop for List<T> {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
