use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer,
        clone,
        sequence_resize,
        sequence_set, sequence_get,
        sequence_length,
    },
};

pub struct Sequence {
    pub(crate) address:Ref,
}
impl Sequence {
    pub fn new() -> Self
    {
        Self { address:unsafe {acquire(Self::type_id())} }
    }

    pub fn with(mut self, data:Vec<u8>) -> Self
    {
        self.set(data);
        self
    }

    pub fn with_str(mut self, data:&str) -> Self
    {
        self.set_str(data);
        self
    }

    pub fn set(&mut self, data:Vec<u8>)
    {
        unsafe {sequence_resize(self.address.clone(), data.len())};
        for i in 0..data.len() {
            unsafe {sequence_set(self.address.clone(), i, data[i])};
        }
    }

    pub fn set_str(&mut self, data:&str)
    {
        let data = data.as_bytes();
        unsafe {sequence_resize(self.address.clone(), data.len())};
        for i in 0..data.len() {
            unsafe {sequence_set(self.address.clone(), i, data[i])};
        }
    }

    pub fn get(&self) -> Vec<u8>
    {
        let mut data = Vec::new();
        for i in 0..data.len() {
            data.push(unsafe {sequence_get(self.address.clone(), i)});
        }
        data
    }

    pub fn get_str(&self) -> String
    {
        let mut data = Vec::new();
        for i in 0..self.len() {
            data.push(unsafe {sequence_get(self.address.clone(), i)});
        }
        String::from_utf8(data).unwrap()
    }

    pub fn set_at(&mut self, index:usize, data:u8)
    {
        unsafe {sequence_set(self.address.clone(), index, data)}
    }

    pub fn get_at(&self, index:usize) -> u8
    {
        unsafe {sequence_get(self.address.clone(), index)}
    }

    pub fn len(&self) -> usize
    {
        unsafe {sequence_length(self.address.clone())}
    }
}
impl SzunType for Sequence {
    fn type_id() -> usize {
        unsafe {type_outer(0, types::TYPE_SEQUENCE)}
    }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}

pub fn sequence(data:Vec<u8>) -> Sequence
{
    Sequence::new().with(data)
}

pub fn string(data:&str) -> Sequence
{
    Sequence::new().with_str(data)
}

impl Clone for Sequence {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())} }
    }
}

impl TryFrom<Ref> for Sequence {
    type Error = SzunError;

    fn try_from(addr: Ref) -> Result<Self, Self::Error> {
        if addr.class == Self::type_id() {
            Ok(Self { address:addr })
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Ref> for Sequence { fn into(self) -> Ref { self.address.clone() } }

impl Drop for Sequence {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
