use crate::{
    prelude::*, types,
    runtime::{
        acquire, release,
        type_outer, type_sanitize,
        clone,
        var_set, var_get,
    },
};

pub struct Var {
    address:Ref,
}
impl Var {
    pub fn new() -> Self {
        Self { address:unsafe {acquire(Self::type_id())} }
    }

    pub fn with<T: SzunType>(mut self, data:&T) -> Self {
        self.set(data);
        self
    }

    pub fn set<T: SzunType>(&mut self, data:&T) {
        unsafe {var_set(self.address.clone(), data.__to_ref())}
    }
    
    pub fn get<T: SzunType>(&self) -> T {
        let refr = unsafe {var_get(self.address.clone())};
        if unsafe {type_sanitize(refr.class)} == T::type_id() {
            T::__from_ref(refr)
        } else { panic!("mismatched szun types"); }
    }

    pub fn try_get<T: SzunType>(&self) -> Result<T,SzunError> {
        let refr = unsafe {var_get(self.address.clone())};
        if unsafe {type_sanitize(refr.class)} == T::type_id() {
            Ok(T::__from_ref(refr))
        } else { Err(SzunError::MismatchedType) }
    }

    pub fn has<T: SzunType>(&self) -> bool {
        let refr = unsafe {var_get(self.address.clone())};
        unsafe {type_sanitize(refr.class) == T::type_id()}
    }

    pub fn clear(&mut self)
    {
        unsafe {var_set(self.address.clone(), Ref::null())}
    }
}
impl SzunType for Var {
    fn type_id() -> usize { unsafe {type_outer(0, types::TYPE_VARYING)} }
    fn to_ref(self) -> Ref { self.address.clone() }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { true }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}

pub fn var<T: SzunType>(data:&T) -> Var
{
    Var::new().with(data)
}

impl Clone for Var {
    fn clone(&self) -> Self {
        Self { address:unsafe {clone(self.address.clone())} }
    }
}

impl From<Ref> for Var {
    fn from(refr: Ref) -> Self {
        if refr.class == Self::type_id() {
            Self { address:refr }
        } else { panic!("mismatched szun types"); }
    }
}
impl Into<Ref> for Var { fn into(self) -> Ref { self.address.clone() } }

impl Drop for Var {
    fn drop(&mut self) {
        unsafe {release(self.address.clone())};
    }
}
